from .utils import expose_static_methods
from pyhectiqlab.step import Step

"""
Functional dataset API. To expose a function, use the `exposed_method` decorator in the `Run` singleton.
"""

locals().update(expose_static_methods(Step))



import clisync
import logging

from typing import Optional, Union, List
import datetime as dt

from pyhectiqlab.project import Project
from pyhectiqlab.run import Run
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias, execute_online_only
from pyhectiqlab.utils import is_running_event_loop

logger = logging.getLogger()

class Message:

    _client: Client = Client
    _cache: dict = {} # Cache str -> Message
    _last_get_all = None
    _delay_get_all = 5 # 5 seconds
    _observed_keys = set()

    def __init__(self, key: str, value: str, run_id: str):
        self.key = key
        self.value = value
        self._run_id = run_id

    @staticmethod
    @execute_online_only
    @functional_alias("publish_message")
    @clisync.include(wait_response=True)
    def publish(
        key: str,
        value: str,
        run: Optional[Union[str, int]] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Publish a message.

        Args:
            key (str): The key of the message.
            value (str): The value of the message.
            run (Union[str, int], optional): The run to publish the message to. Defaults to None.
            project (str, optional): The project to publish the message to. Defaults to None.
            wait_response (bool, optional): Whether to wait for a response. Defaults to False.
        """

        if not key:
            logger.error("A key must be provided.")
            return
        if not value:
            logger.error("A value must be provided.")
            return

        run_id = None
        project = Project.get(project)
        if isinstance(run, int):
            if not project:
                logger.error("A project must be provided when `run` is an integer.")
                return
            run = Run.retrieve(rank=run, project=project)
            if not run:
                logger.error(f"Run with rank {run} not found.")
                return
            run_id = run["id"]
        else:
            run_id = Run.get_id(run)
        
        if not run_id:
            logger.error("A run must be provided.")
            return
        
        
        return Message._client.post("/app/messages/publish", 
                                    json={"key": key, "value": value, "run": run_id}, 
                                    wait_response=wait_response)

    @staticmethod
    @execute_online_only
    @functional_alias("get_all_messages")
    # @clisync.include(wait_response=True)
    def get_all(
        run: Optional[Union[str, int]] = None,
        project: Optional[str] = None,
        wait_response: bool = True,
    ):
        """Get all messages for a run.

        Args:
            run (Union[str, int], optional): The run to get messages for. Defaults to None.
            project (str, optional): The project to get messages for. Defaults to None.

        Returns:
            List[Message]: A list of messages if `wait_response` is True. Otherwise, None.
          
        """
        if Message._last_get_all:
            diff = dt.datetime.now() - Message._last_get_all
            if diff.total_seconds() < Message._delay_get_all:
                logger.debug(f"Message.get_all() called too soon. Ignoring. Waited {diff.total_seconds()} seconds.")
                return

        Message._last_get_all = dt.datetime.now()

        if is_running_event_loop():
            logger.error("Messages cannot be fetched from an event loop (jupyter notebook). Future support will be added.")
            return

        project = Project.get(project)
        if isinstance(run, int):
            if not project:
                logger.error("A project must be provided when `run` is an integer.")
                return
            run = Run.retrieve(rank=run, project=project)
            if not run:
                logger.error(f"Run with rank {run} not found.")
                return
            run_id = run["id"]
        else:
            run_id = Run.get_id(run)
        
        if not run_id:
            logger.error("A run must be provided.")
            return
        if wait_response:
            results = Message._client.get("/app/messages", 
                                        params={"run": run_id}, 
                                        wait_response=True)
            results  = results or {}
            return [Message(key, value, run_id) for key, value in results.items()]
        else:
            def composition():
                results = Message._client.get("/app/messages", 
                                            params={"run": run_id}, 
                                            wait_response=True)
                for key, value in results.items():
                    # Cache the message
                    Message._cache[key] = Message(key, value, run_id)

            Message._client.execute(composition,
                                    wait_response=False, 
                                    is_async_method=False)
    
    def ack(self):
        """Acknowledge a message.
        """
        if not self.key:
            logger.error("`ack()` can only be called on a message object with `self.key`.")
            return
        if not self._run_id:
            logger.error("`ack()` can only be called on a message object with a self._run_id.")
            return
        # Remove from _cache
        if self.key in Message._cache:
            Message._cache.pop(self.key, None)
        return Message._client.post("/app/messages/ack", 
                                    json={"key": self.key, "run": self._run_id}, 
                                    wait_response=True)

    @staticmethod
    @execute_online_only
    @functional_alias("get_message")
    def get(key: str,
            run_id: Optional[str] = None):
        """Method to get a message by key. The message is retrieved from the cache 
        if it exists. When first called, it will fetch all messages for the active
        run.

        Args:
            key (str): The key of the message.
            run_id (Optional[str], optional): The run to get the message from. Defaults to None.
        """
        if not key:
            logger.error("A key must be provided.")
            return
        
        if key not in Message._observed_keys:
            Message._observed_keys.add(key)
            Message.declare_keys([key])

        Message.get_all(run=run_id, wait_response=False) # Warm up the cache
        return Message._cache.get(key)
    
    @staticmethod
    @execute_online_only
    @functional_alias("declare_observed_messages")
    def declare_keys(keys: List[str], 
                     run_id: Optional[str] = None,
                     wait_response: bool = False):
        """Declare the keys observed by the client.

        Args:
            keys (list): A list of keys observed by the client.
        """
        run_id = Run.get_id(run_id)
        return Message._client.post("/app/messages/declare", 
                                    json={"keys": keys, "run": run_id}, 
                                    wait_response=wait_response)

    def __str__(self):
        if not self.key:
            return "Message()"
        return f"Message(key={self.key}, value={self.value})"
    
    def __repr__(self):
        return str(self)


import os
import httpx
import toml
import logging

from pyhectiqlab import API_URL
from pyhectiqlab.settings import getenv
from pyhectiqlab.logging import hectiqlab_logger as logger
from pathlib import Path
from typing import Optional, Union


def load_credentials(name: Optional[str] = None, display_error: Optional[bool] = True) -> Union[None, str]:
    path = getenv("HECTIQLAB_CREDENTIALS", os.path.join(Path.home(), ".hectiq-lab", "credentials.toml"))

    logger.info(f"⏳ Loading credentials from {path}...")
    if not os.path.exists(path):
        if display_error:
            logger.error(f"❌ Credentials file not found at {path}.")
        return None

    with open(path, "r") as path:
        data = toml.load(path)
    # If key_name is provided, we return the value of the key_name
    if name is not None:
        return data[name]
    # Otherwise, we return the first value that is not None
    if len(data) > 0:
        return next(iter(data.values()))
    if display_error:
        logger.error("❌ No API key found in the credentials file.")
    return


def is_authenticated(name: Optional[str] = None, display_error: Optional[bool] = True) -> bool:
    """Checks if local is authenticated to the Hectiq Lab.
    If API key is None, the api_key is taken from:
     - the api_key located in the file `~/.hectiq-lab/credentials.toml`
     (or the HECTIQLAB_CREDENTIALS).

    Args:
        name (Optional[str], optional): Name of the API key. If None, the first available key is used.
            Default: None.

    Returns:
        bool: True if the user is authenticated, False otherwise.
    """

    data = load_credentials(name, display_error=display_error)
    if data is None:
        return False
    key = data.get("value")
    if key is None:
        return False
    return True


class Auth(httpx.Auth):
    """Authentication class for the Hectiq Lab API. This class is used to authenticate
    via an access token that is refreshed when it expires."""

    def __init__(self, name: Optional[str] = None):
        super().__init__()
        self.data = load_credentials(name)
        self.access_token = None
        self.refresh_token()

    def is_logged(self):
        return is_authenticated()

    @property
    def online(self):
        return self.access_token is not None

    @property
    def user(self):
        if self.data is None:
            return None
        return self.data.get("user")

    @property
    def api_key(self):

        if getenv("HECTIQLAB_API_KEY") is not None:
            return getenv("HECTIQLAB_API_KEY")
        if self.data is None:
            return None
        return self.data.get("value")

    @property
    def auth_header(self):
        return {"X-API-Key": self.api_key} if self.api_key else {}

    @property
    def auth_token_header(self):
        return {"Authentification": f"Bearer {self.access_token}"} if self.access_token is not None else {}

    def refresh_token(self):
        if self.api_key is None:
            logging.error("API key is not provided. Continuing in offline mode.")
            return
        response = httpx.post(f"{API_URL}/app/auth/auth-api-key", headers=self.auth_header)
        if response.status_code != 200:
            logging.error("User could not be authenticated. Continuing in offline mode.")
            return
        if "access_token" in response.cookies:
            self.access_token = response.cookies["access_token"]
        else:
            logging.error("Access token missing. User could not be authenticated. Continuing in offline mode.")
            return

    def auth_flow(self, request: httpx.Request):
        request.headers.update(self.auth_token_header)
        response = yield request

        if response.status_code == 401 or response.status_code == 403:
            response = yield self.build_refresh_request()
            if "access_token" in response.cookies:
                self.access_token = response.cookies["access_token"]
            else:
                logging.error("Access token missing. User could not be authenticated. Continuing in offline mode.")
                return
            request.headers.update(self.auth_token_header)
            response = yield request

    def build_refresh_request(self):
        return httpx.Request("POST", f"{API_URL}/app/auth/auth-api-key", headers=self.auth_header)



import pytest
import os
import logging

from utils import random_uuid

from pyhectiqlab.project import Project

logger = logging.getLogger()

Project.set("hectiq-ai/test")


@pytest.fixture
def path():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "dummy/"))


@pytest.fixture
def save_path():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))


@pytest.fixture
def clear_data(data_path):
    yield
    import shutil

    shutil.rmtree(data_path, ignore_errors=True)


def test_model_create_basic(path):
    from pyhectiqlab import Model

    model = Model.create(name="test", path=path)
    logger.info(f"Model created: {model}")


def test_model_create_cli(path):
    import os

    project = Project.get()
    os.system(f"hectiq-lab Model.create --name test --path {path} --project {project}")


def test_model_create_with_run(path):
    from pyhectiqlab import Model, Run

    run = Run(title=f"test {random_uuid()}")
    model = Model.create(name="test", path=path)
    logger.info(f"Model created with run: {model}")


def test_model_create_and_upload(path):
    from pyhectiqlab import Model

    model = Model.create(name="test", path=path, project="hectiq-ai/test", upload=True)
    logger.info(f"Model created and uploaded: {model}")


def test_model_create_and_upload_cli(path):
    import os

    project = Project.get()
    os.system(f"hectiq-lab Model.create --name test --path {path} --project {project} --upload")


def test_download(save_path):
    from pyhectiqlab import Model

    model = Model.download(name="test", version="1.2", project="hectiq-ai/test", path=save_path)
    logger.info(f"Model downloaded: {model}")


def test_download_with_run(save_path):
    from pyhectiqlab import Model, Run

    run = Run(title=f"test {random_uuid()}")
    from pyhectiqlab.run import Run

    run_id = Run.get_id()
    logger.info(f"Run ID: {run_id}")
    model = Model.download(name="test", version="1.2", project="hectiq-ai/test", path=save_path)
    logger.info(f"Model downloaded with run: {model}")


def test_download_cli(save_path):
    import os

    project = Project.get()
    os.system(f"hectiq-lab Model.download --name test --version 1.2 --project {project} --path {save_path} ")


def test_model_retrieve():
    from pyhectiqlab import Model

    model = Model.retrieve(name="test-model", version="1.3", project="hectiq-ai/test")
    assert model["name"] == "test-model"
    assert model["version"] == "1.3"


def test_model_add_tag():
    from pyhectiqlab import Model

    Model.add_tags(tags=["test", "model"], name="test-model", version="1.3", project="hectiq-ai/test")


if __name__ == "__main__":
    pass



import pytest
import numpy as np

from pyhectiqlab import Run


def test_push_metrics():
    run = Run(31, project="hectiq-ai/test")

    for i in range(500):
        run.add_metric("metric1", step=i, value=np.random.rand())



import os
import clisync

from typing import Optional, List

from pyhectiqlab.project import Project
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias, on_accelerator_main_process
from pyhectiqlab.utils import is_running_event_loop
from pyhectiqlab.logging import hectiqlab_logger as logger


class Artifact:
    _client: Client = Client

    @staticmethod
    @clisync.include(wait_response=True)
    @on_accelerator_main_process()
    def create(
        path: str,
        run_id: str,
        project: str,
        name: Optional[str] = None,
        step: Optional[int] = None,
        group: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Upload an artifact to a run.

        Args:
            path (str): Path to the file to upload.
            run_id (str): ID of the run.
            project (str): ID of the project.
            name (str, optional): Name of the artifact. If None, the basename of the file is used. Default: None.
            step (int, optional): Step number. If None, the artifact is not considered as a step artifact. Default: None.
            group (str, optional): Group name. If None, the group is determined from the path. Default: None.
            wait_response (bool): Set to true to upload sync. If False, the upload is made in background. Default: False.

        Raises:
            logger.error: If `path`, `run_id` and `project` are not provided, it does not upload the artifact and returns None.
        """

        if path is None or run_id is None or project is None:
            logger.error("Parameters `path`, `run_id` and `project` are required.")
            return
        name = name or os.path.basename(path)
        group = group or os.path.basename(path)
        num_bytes = os.path.getsize(path)
        extension = os.path.splitext(path)[1]

        json = {
            "name": name,
            "num_bytes": num_bytes,
            "group": group,
            "step": step,
            "project": project,
            "run": run_id,
            "extension": extension,
        }

        def composition_sync(**kwargs):
            artifact = Client.post("/app/artifacts", wait_response=True, json=json)
            if not artifact:
                return
            return Client.upload_sync(local_path=path, policy=artifact.get("upload_policy"), **kwargs)

        async def composition_async(**kwargs):
            artifact = Client.post("/app/artifacts", wait_response=True, json=json)
            if not artifact:
                return
            return await Client.upload_async(local_path=path, policy=artifact.get("upload_policy"), **kwargs)

        if is_running_event_loop():
            Client.execute(composition_sync, wait_response=wait_response, is_async_method=False, with_progress=True)
        else:
            Client.execute(composition_async, wait_response=wait_response, is_async_method=True, with_progress=True)
        return path

    @staticmethod
    @functional_alias("delete_artifact")
    @clisync.include(wait_response=True)
    @on_accelerator_main_process()
    def delete(
        id: str,
        wait_response: bool = False,
    ):
        """
        Delete an artifact.

        Args:
            id (str): ID of the artifact.
            wait_response (bool): Set to true to delete sync. If False, the deletion is made in background. Default: False.
        """
        return Client.delete(f"/app/artifacts/{id}", wait_response=wait_response)

    @staticmethod
    @functional_alias("get_artifact")
    @clisync.include()
    @on_accelerator_main_process()
    def retrieve(id: str, fields: Optional[List[str]] = None):
        """
        Retrieve an artifact.

        Args:
            id (str): ID of the artifact.
        """
        return Client.get(f"/app/artifacts/{id}", wait_response=True, params={"fields": fields})

    @staticmethod
    @functional_alias("download_artifact")
    @clisync.include()
    @on_accelerator_main_process()
    def download(
        id: str,
        path: str = "./",
    ):
        """
        Download an artifact from a run.

        Args:
            id (str): Id of the artifact.
            path (str): Path to save the file. Default: './'.
        """
        path = path or "./"

        def composition_sync(**kwargs):
            fields = ["download_url", "num_bytes", "name"]
            artifact = Client.get(f"/app/artifacts/{id}", wait_response=True, params={"fields": fields})
            if not artifact:
                return
            os.makedirs(path, exist_ok=True)
            Client.download_sync(
                url=artifact["download_url"],
                local_path=os.path.join(path, artifact["name"]),
                num_bytes=artifact["num_bytes"],
                **kwargs,
            )

        async def composition_async(**kwargs):
            fields = ["download_url", "num_bytes", "name"]
            artifact = Client.get(f"/app/artifacts/{id}", wait_response=True, params={"fields": fields})
            if not artifact:
                return
            os.makedirs(path, exist_ok=True)
            await Client.download_async(
                url=artifact["download_url"],
                local_path=os.path.join(path, artifact["name"]),
                num_bytes=artifact["num_bytes"],
                **kwargs,
            )

        if is_running_event_loop():
            Client.execute(composition_sync, wait_response=True, is_async_method=False, with_progress=True)
        else:
            Client.execute(composition_async, wait_response=True, is_async_method=True, with_progress=True)
        return path

    @staticmethod
    @functional_alias("list_artifacts")
    @clisync.include()
    def list(
        run_id: str,
        project: Optional[str] = None,
        fields: Optional[List[str]] = None,
        group_by_step: Optional[bool] = True,
        page: Optional[int] = 1,
        limit: Optional[int] = 50,
        order_by: Optional[str] = "name",
        order_direction: Optional[str] = "asc",
    ):
        """
        List the artifacts of a run.

        Args:
            run_id (str): ID of the run.
            project (str): ID of the project.
            fields (List[str], optional): List of fields to retrieve. Default: None.
            group_by_step (bool, optional): Set to True to group the artifacts by step. Default: True.
            page (int, optional): Page number. Default: 1.
            limit (int, optional): Number of artifacts to retrieve per page. Default: 50.
            order_by (str, optional): Field to order by. Default: "name".
            order_direction (str, optional): Order direction. Default: "asc".
        """
        project = Project.get(project)
        if not project:
            return

        return Client.get(
            "/app/artifacts",
            wait_response=True,
            params={
                "run": run_id,
                "project": project,
                "fields": fields,
                "keep_latest_step": True if group_by_step else False,
                "page": page,
                "limit": limit,
                "order_by": order_by,
                "order_direction": order_direction,
            },
        )
